//
//  Application program: Test of USB-Bluetooth adapter
//

#include <time.h>
#include <string.h>

// Harmony
#include "usb/usb_host.h"               // Usb host class common
#include "usb/usb_host_bluetooth.h"     // Usb host Wireless controler class + bluetooth-HCI sub class

// Bluetooth stack
#include "bt_rfcomm_server.h"           // Bluetooth RFCOMM API
#include "bt_att_server.h"              // Bluetooth ATT server API
#include "bt_att_db.h"                  // Bluetooth ATT database API

// Local
#include "rtc.h"                        // time()
#include "app.h"
#include "bsp_config.h"


//
// Application data
//
typedef enum {
    APP_STATE_OPEN_HOST_LAYER = 0,
    APP_STATE_WAIT_FOR_HOST_ENABLE,
    APP_STATE_WAIT_FOR_DEVICE_ATTACH,
    APP_STATE_DEVICE_CONNECTED,
    APP_STATE_IDLE,
    APP_STATE_DEVICE_DISCONNECTED,
    APP_STATE_ERROR
} APP_STATES;

typedef struct {
    USB_HOST_HANDLE hostHandle;
    APP_STATES      state;
    time_t          tm;
    char           *msg;
} APP_DATA;

APP_DATA appData;


//
// USB host callback routine
//
USB_HOST_EVENT_RESPONSE APP_UsbHostEvent(USB_HOST_EVENTS event, void *eventData, uintptr_t context){
    switch(event) {
    case USB_HOST_EVENT_VBUS_REQUEST_POWER:         // N/A
        break;
    case USB_HOST_EVENT_UNSUPPORTED_DEVICE:         // N/A
        break;
    case USB_HOST_EVENT_CANNOT_ENUMERATE:           // Unsupported device descriptor
        break;
    case USB_HOST_EVENT_CONFIGURATION_COMPLETE:     // N/A
        break;
    case USB_HOST_EVENT_CONFIGURATION_FAILED:       // N/A
        break;
    case USB_HOST_EVENT_DEVICE_SUSPENDED:           // N/A
        break;
    case USB_HOST_EVENT_DEVICE_RESUMED:             // N/A
        break;
    }
    return(USB_HOST_EVENT_RESPONSE_NONE);
}


//
//  USB host <class> call back routine
//
bool APP_UsbHostBluetoothEvent(USB_HOST_BLUETOOTH_INDEX index, USB_HOST_BLUETOOTH_EVENT event, void* pData){
    switch(event) {
    case USB_HOST_BLUETOOTH_EVENT_ATTACH:
        appData.state = APP_STATE_DEVICE_CONNECTED;
        break;

    case USB_HOST_BLUETOOTH_EVENT_DETACH:
        appData.state = APP_STATE_DEVICE_DISCONNECTED;
        break;

    default:
        break;
    }
    return(0);
}


//
//  Bluetooth call back routines
//
void bt_rfcommServerDataRecived(unsigned char *inBuff, size_t inLen){
    // Sample code
    //printf("RFCOMM recv: %s", inBuff);
    return;
}

void bt_attDbNotification(unsigned short handle, void *value, size_t valueLen){
    bt_att_serverNotificationQueueAdd(handle);  // Essential action
    switch(handle) {                            // User action
    case 0x1002:
        if (*(unsigned char *)value == 0x00) {
            BSP_LEDOff(BSP_LED_2);
        } else {
            BSP_LEDOn(BSP_LED_2);
        }
        break;
    default:
        break;
    }
    return;
}


//
// Application
//
void APP_Initialize(void){
    appData.state = APP_STATE_OPEN_HOST_LAYER;
    appData.msg   = "Hello world.\n";
    return;
}

void APP_Tasks(void){
    switch(appData.state) {
    case APP_STATE_OPEN_HOST_LAYER:
        appData.hostHandle = USB_HOST_Open(USB_HOST_INDEX_0, DRV_IO_INTENT_EXCLUSIVE);
        if (appData.hostHandle != USB_HOST_HANDLE_INVALID) {
            USB_HOST_OperationEnable(appData.hostHandle);
            appData.state = APP_STATE_WAIT_FOR_HOST_ENABLE;
        }
        break;

    case APP_STATE_WAIT_FOR_HOST_ENABLE:
        if(USB_HOST_OperationIsEnabled(appData.hostHandle)) {
            USB_HOST_EventCallBackSet(appData.hostHandle, APP_UsbHostEvent, 0);
            USB_HOST_BLUETOOTH_EventHandlerSet(APP_UsbHostBluetoothEvent);
            rtc_open();
            appData.state = APP_STATE_WAIT_FOR_DEVICE_ATTACH;
        }
        break;

    case APP_STATE_WAIT_FOR_DEVICE_ATTACH:
        // Wait for device attach. The state machine will move
        // to the next state when the attach event is received.
        break;

    case APP_STATE_DEVICE_CONNECTED:
        BSP_LEDOn(BSP_LED_2);   // Plugged the bluetooth dongle
        appData.state = APP_STATE_IDLE;
        break;

    case APP_STATE_IDLE:
        if (appData.tm != time(NULL)) {                                 // 1sec interval
            appData.tm = time(NULL);
            bt_rfcommServerDataSendRequest((unsigned char *)appData.msg, strlen(appData.msg));
            bt_attDbUpdate(0x1012, &appData.tm, sizeof(appData.tm));    // Attribute protocol
        }
        break;

    case APP_STATE_DEVICE_DISCONNECTED:
        BSP_LEDOff(BSP_LED_2);   // Un plugged the bluetooth dongle
        appData.state = APP_STATE_WAIT_FOR_DEVICE_ATTACH;
        break;

    case APP_STATE_ERROR:
        for(;;);    // STOP!!
        break;

    default:
        break;
    }
    return;
}
