//
//  bt_security_manager.c
//
//  Bluetooth Protocol Stack - Security Manager (Just Works Pairing Method)
//  Copyright (C) 2014 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//          [Vol 3] Core System Package [Host volume]
//           Part H: SECURITY MANAGER SPECIFICATION
//      Federal Information Processing Standards Publication 197 (FIPS-197)
//          ADVANCED ENCRYPTION STANDARD (AES)
//          Rijndael Algorithm
//              http://csrc.nist.gov/archive/aes/rijndael/wsdindex.html
//              Reference source code
//                  rijndael-alg-ref.[ch]
//                  rijndael-api-ref.[ch]
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2014.02.23 v0.0  First cording
//  2014.03.05 v0.2  Commit
//

#include <stddef.h>
#include <string.h>
#include "bt_spec_smp.h"
#include "bt_l2cap_cb_common.h"
#include "bt_l2cap_cb_link.h"
#include "rijndael-api-ref.h"           // AES cryptographic routine

#define MEMCPY(d,s)     memcpy(d, s, sizeof(d))
#define MEMCMP(d,s)     memcmp(d, s, sizeof(d))
#define MEMSET(d,s)     memset(d, s, sizeof(d))
//#define DEBUG

// DEBUG tools
#ifdef DEBUG
#include <stdio.h>                      // Debug
#define DEBUGP(...)     printf(__VA_ARGS__)
#define DUMP(t,p)       bt_securityManagerDump(t,p,sizeof(p))
void bt_securityManagerDump(char *t, unsigned char *p, int l){
    int i;

    printf(t);
    for(i = 0; i < l; i++) {
        if (((i % 4) == 0) && (i > 0)) {
            printf(",");
        }
        printf("%02x", *p++);
    }
    printf("\n");
}
#else
#define DEBUGP(...)
#define DUMP(...)
#endif

// Security function e
void bt_securityManagerCrypt_convCopy(unsigned char *out, unsigned char *in){
    out[ 0] = in[15];
    out[ 1] = in[14];
    out[ 2] = in[13];
    out[ 3] = in[12];
    out[ 4] = in[11];
    out[ 5] = in[10];
    out[ 6] = in[ 9];
    out[ 7] = in[ 8];
    out[ 8] = in[ 7];
    out[ 9] = in[ 6];
    out[10] = in[ 5];
    out[11] = in[ 4];
    out[12] = in[ 3];
    out[13] = in[ 2];
    out[14] = in[ 1];
    out[15] = in[ 0];
    return;
}
void bt_securityManagerCrypt_e(unsigned char *encorded, unsigned char *key, unsigned char *plaintextData){
    keyInstance     ki;
    cipherInstance  ci;
    char            keyStr[33];
    int             i;
    int             rc;
    unsigned char   in[16];
    unsigned char   out[16];

    //
    //  rijndael API convert
    //
    bt_securityManagerCrypt_convCopy(in, plaintextData);
    memset(&ki, 0, sizeof(ki));
    memset(&ci, 0, sizeof(ci));
    for(i = 0; i < 16; i++) {
        sprintf(&(keyStr[i*2]), "%02x", key[15 - i]);
    }
    keyStr[i*2] = '\0';
    ki.blockLen = 128;
    rc = makeKey(&ki, DIR_ENCRYPT, 128, keyStr);
    if (rc == FALSE) {
        DEBUGP("makeKey error=%d\n", rc);
    }
    ci.blockLen = 128;
    rc = cipherInit(&ci, MODE_ECB, "00000000000000000000000000000000");
    if (rc == FALSE) {
        DEBUGP("cipherInit error=%d\n", rc);
    }
    rc = blockEncrypt(&ci, &ki, in, 128, out);
//  DEBUGP("blockEncrypt error=%d\n", rc);
//  DUMP("txt=", in);
//  DUMP("enc=", out);
    bt_securityManagerCrypt_convCopy(encorded, out);
    return;
}

// Random Address Hash function ah
void bt_securityManagerCrypt_ah(unsigned char *out, unsigned char *k, unsigned char *r){
    unsigned char   rd[16];

    rd[15] = 0;
    rd[14] = 0;
    rd[13] = 0;
    rd[12] = 0;
    rd[11] = 0;
    rd[10] = 0;
    rd[ 9] = 0;
    rd[ 8] = 0;
    rd[ 7] = 0;
    rd[ 6] = 0;
    rd[ 5] = 0;
    rd[ 4] = 0;
    rd[ 3] = 0;
    rd[ 2] = r[2];
    rd[ 1] = r[1];
    rd[ 0] = r[0];
    bt_securityManagerCrypt_e(out, k, rd);
    out[15] = 0;
    out[14] = 0;
    out[13] = 0;
    out[12] = 0;
    out[11] = 0;
    out[10] = 0;
    out[ 9] = 0;
    out[ 8] = 0;
    out[ 7] = 0;
    out[ 6] = 0;
    out[ 5] = 0;
    out[ 4] = 0;
    out[ 3] = 0;
    return;
}

// Confirm value generation function c1
void bt_securityManagerCrypt_c1(unsigned char *out,
        unsigned char *k, unsigned char *r, unsigned char *preq, unsigned char *pres, int iat, unsigned char *ia, int rat, unsigned char *ra){
    unsigned char   p1[16];
    unsigned char   p2[16];
    unsigned char   px[16];
    int             i;
//  unsigned char   p1_test[16] = {0x05,0x00,0x08,0x00,0x00,0x03,0x02,0x07,0x07,0x10,0x00,0x00,0x01,0x01,0x00,0x01};
//  unsigned char   p2_test[16] = {0x00,0x00,0x00,0x00,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6};

    p1[15] = pres[6];
    p1[14] = pres[5];
    p1[13] = pres[4];
    p1[12] = pres[3];
    p1[11] = pres[2];
    p1[10] = pres[1];
    p1[ 9] = pres[0];
    p1[ 8] = preq[6];
    p1[ 7] = preq[5];
    p1[ 6] = preq[4];
    p1[ 5] = preq[3];
    p1[ 4] = preq[2];
    p1[ 3] = preq[1];
    p1[ 2] = preq[0];
    p1[ 1] = rat;
    p1[ 0] = iat;
    p2[15] = 0;
    p2[14] = 0;
    p2[13] = 0;
    p2[12] = 0;
    p2[11] = ia[5];
    p2[10] = ia[4];
    p2[ 9] = ia[3];
    p2[ 8] = ia[2];
    p2[ 7] = ia[1];
    p2[ 6] = ia[0];
    p2[ 5] = ra[5];
    p2[ 4] = ra[4];
    p2[ 3] = ra[3];
    p2[ 2] = ra[2];
    p2[ 1] = ra[1];
    p2[ 0] = ra[0];
//  bt_securityManagerCrypt_convCopy(p1, p1_test);
//  bt_securityManagerCrypt_convCopy(p2, p2_test);
//  DUMP("p1=", p1);
//  DUMP("p2=", p2);
    for(i = 0; i < 16; i++) {
        p1[i] ^= r[i];
    }
    bt_securityManagerCrypt_e(px, k, p1);
    for(i = 0; i < 16; i++) {
        px[i] ^= p2[i];
    }
    bt_securityManagerCrypt_e(out, k, px);
    return;
}

// Key generation function s1
void bt_securityManagerCrypt_s1(unsigned char *out, unsigned char *k, unsigned char *r1, unsigned char *r2){
    unsigned char   rd[16];

    rd[15] = r1[ 7];
    rd[14] = r1[ 6];
    rd[13] = r1[ 5];
    rd[12] = r1[ 4];
    rd[11] = r1[ 3];
    rd[10] = r1[ 2];
    rd[ 9] = r1[ 1];
    rd[ 8] = r1[ 0];
    rd[ 7] = r2[ 7];
    rd[ 6] = r2[ 6];
    rd[ 5] = r2[ 5];
    rd[ 4] = r2[ 4];
    rd[ 3] = r2[ 3];
    rd[ 2] = r2[ 2];
    rd[ 1] = r2[ 1];
    rd[ 0] = r2[ 0];
    bt_securityManagerCrypt_e(out, k, rd);
    return;
}

// 5.1.1.1 DIV Mask generation function dm
void bt_securityManagerCrypt_dm(unsigned char *out, unsigned char *k, unsigned char *r){
    unsigned char   rd[16];

    rd[15] = 0;
    rd[14] = 0;
    rd[13] = 0;
    rd[12] = 0;
    rd[11] = 0;
    rd[10] = 0;
    rd[ 9] = 0;
    rd[ 8] = 0;
    rd[ 7] = r[ 7];
    rd[ 6] = r[ 6];
    rd[ 5] = r[ 5];
    rd[ 4] = r[ 4];
    rd[ 3] = r[ 3];
    rd[ 2] = r[ 2];
    rd[ 1] = r[ 1];
    rd[ 0] = r[ 0];
    bt_securityManagerCrypt_e(out, k, rd);
    out[15] = 0;
    out[14] = 0;
    out[13] = 0;
    out[12] = 0;
    out[11] = 0;
    out[10] = 0;
    out[ 9] = 0;
    out[ 8] = 0;
    return;
}

//
//  Server
//
size_t bt_securityManager(unsigned short handle, unsigned char *inBuff, size_t inLen, unsigned char *outBuff, size_t outSize) {
    struct bt_smp_PDU_Format    *in;
    struct bt_smp_PDU_Format    *out;
    size_t                      len;
    struct bt_l2capLink         *cb;
    static unsigned char        tk[16]    = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    static unsigned char        srand[16] = {0xE0,0x2E,0x70,0xC6,0x4E,0x27,0x88,0x63,0x0E,0x6F,0xAD,0x56,0x21,0xD5,0x83,0x57};

    in  = (struct bt_smp_PDU_Format *)inBuff;
    out = (struct bt_smp_PDU_Format *)outBuff;
    len = 0;
    cb  = bt_l2capLinkFind(NULL, handle);   // Not NULL

    switch(cb->smState) {
    case BT_L2CAP_LINK_SM_IDLE:

    case BT_L2CAP_LINK_SM_PHASE1:
        //
        // Phase 1
        //
        if (in != NULL) {
            switch(in->Code) {
            case SMP_Code_Pairing_Request:
                DEBUGP("SM: Pairing_Request io=0x%02x auth=%d OOB=%d\n",
                        in->Pairing_Request.IO_Capability,
                        in->Pairing_Response.AuthReq,
                        in->Pairing_Request.OOB_data_flag);
                MEMCPY(cb->preq, in);
                out->Code = SMP_Code_Pairing_Response;
                out->Pairing_Response.IO_Capability = SMP_IO_Capability_NoInputNoOutput;
//              out->Pairing_Response.AuthReq       = SMP_AuthReq_No_Bonding;
                out->Pairing_Response.AuthReq       = SMP_AuthReq_Bonding;
                out->Pairing_Response.OOB_data_flag = SMP_OOB_data_flag_Authentication_data_not_present;
                out->Pairing_Response.Maximum_Encryption_Key_Size = 16;
//              out->Pairing_Response.Initiator_Key_Distribution  = SMP_Key_Distribution_EncKey;
                out->Pairing_Response.Initiator_Key_Distribution  = 0;
                out->Pairing_Response.Responder_Key_Distribution  = SMP_Key_Distribution_EncKey;
                len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Response);
                MEMCPY(cb->pres, out);
                DEBUGP("SM: Pairing_Response send\n");
                // Create sconfirm
                bt_securityManagerCrypt_c1(cb->sconfirm, tk, srand, cb->preq, cb->pres, cb->device->peerAddressType, cb->device->bdAddr, bt_l2capCommon.myAddressType, bt_l2capCommon.myBdAddr);
                cb->smState = BT_L2CAP_LINK_SM_PHASE2;
                break;
            case SMP_Code_Pairing_Failed:
                DEBUGP("SM: Phase1 Faild!! Reason=0x%02x\n", in->Pairing_Failed.Reason);
                cb->smState = BT_L2CAP_LINK_SM_IDLE;
                break;
            default:
                DEBUGP("SM: Unsupport Code=0x%02x phase1\n", in->Code);
                out->Code                  = SMP_Code_Pairing_Failed;
                out->Pairing_Failed.Reason = SMP_Reason_Command_Not_Supported;
                len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Failed);
                break;
            }
        }
        break;

    case BT_L2CAP_LINK_SM_PHASE2:
        //
        // Phase 2
        //
        if (in != NULL) {
            switch(in->Code) {
            case SMP_Code_Pairing_Confirm:
                DEBUGP("SM: Pairing_Confirm\n");
                MEMCPY(cb->mconfirm, in->Pairing_Confirm.Confirm_value);
                MEMCPY(out->Pairing_Confirm.Confirm_value, cb->sconfirm);
                len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Confirm);
                break;
            case SMP_Code_Pairing_Random:
            {
                unsigned char   chk[sizeof(in->Pairing_Confirm.Confirm_value)];

                DEBUGP("SM: Pairing_Random\n");
                MEMCPY(cb->mrand, in->Pairing_Random.Random_value);
                bt_securityManagerCrypt_c1(chk, tk, cb->mrand, cb->preq, cb->pres, cb->device->peerAddressType, cb->device->bdAddr, bt_l2capCommon.myAddressType, bt_l2capCommon.myBdAddr);
                if (MEMCMP(chk, cb->mconfirm)) {
                    DEBUGP("SM: mrand unmatch!\n");
                    out->Code                  = SMP_Code_Pairing_Failed;
                    out->Pairing_Failed.Reason = SMP_Reason_Confirm_Value_Failed;
                    len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Failed);
                } else {
                    DEBUGP("SM: mrand ok\n");
                    MEMCPY(out->Pairing_Random.Random_value, srand);
                    len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Random);
                    // Create STK
                    bt_securityManagerCrypt_s1(cb->shortTermKey, tk, srand, cb->mrand);
                    cb->smState = BT_L2CAP_LINK_SM_PHASE3;
                }
                break;
            }
            case SMP_Code_Pairing_Failed:
                DEBUGP("SM: Phase2 Faild!! Reason=0x%02x\n", in->Pairing_Failed.Reason);
                cb->smState = BT_L2CAP_LINK_SM_IDLE;
                break;
            default:
                DEBUGP("SM: Unsupport Code=0x%02x\n", in->Code);
                out->Code                  = SMP_Code_Pairing_Failed;
                out->Pairing_Failed.Reason = SMP_Reason_Command_Not_Supported;
                len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Failed);
                break;
            }
        }
        break;

    case BT_L2CAP_LINK_SM_PHASE3:
        //
        // Phase 3
        //
        if (cb->encryption) {
            DEBUGP("SM: Phase3 encryption on\n");
            cb->smState = BT_L2CAP_LINK_SM_PHASE3_SEND_LTK;
//          cb->smState = BT_L2CAP_LINK_SM_PHASE3_IDLE;
        }
        break;
    case BT_L2CAP_LINK_SM_PHASE3_SEND_LTK:
        // Create LTK,EDIV,Rand
        {
            unsigned char   y[16];
            unsigned char   div[2];

            MEMSET(cb->longTermKey,           0x11);
            MEMSET(cb->randomNumber,          0x33);
            MEMSET(div,                       0x55);

            bt_securityManagerCrypt_dm(y, cb->longTermKey, cb->randomNumber);
            cb->encryptionDiversifier[0] = div[0] ^ y[0];
            cb->encryptionDiversifier[1] = div[1] ^ y[1];
        }
        DEBUGP("SM: Encryption_Information send\n");
        out->Code = SMP_Code_Encryption_Information;
        MEMCPY(out->Encryption_Information.Long_Term_Key, cb->longTermKey);
        len = SMP_PDU_Format_HeadSize + sizeof(out->Encryption_Information);
        cb->smState = BT_L2CAP_LINK_SM_PHASE3_SEND_EDIV;
        break;
    case BT_L2CAP_LINK_SM_PHASE3_SEND_EDIV:
        DEBUGP("SM: Master_Identification send\n");
        out->Code = SMP_Code_Master_Identification;
        MEMCPY(out->Master_Identification.EDIV, cb->encryptionDiversifier);
        MEMCPY(out->Master_Identification.Rand, cb->randomNumber);
        len = SMP_PDU_Format_HeadSize + sizeof(out->Master_Identification);
//      cb->smState = BT_L2CAP_LINK_SM_PHASE3_SEND_ID;
        cb->smState = BT_L2CAP_LINK_SM_PHASE3_IDLE;
        break;
    case BT_L2CAP_LINK_SM_PHASE3_SEND_ID:
        cb->smState = BT_L2CAP_LINK_SM_PHASE3_SEND_ADDR;
        break;
    case BT_L2CAP_LINK_SM_PHASE3_SEND_ADDR:
        cb->smState = BT_L2CAP_LINK_SM_PHASE3_SEND_SIG;
        break;
    case BT_L2CAP_LINK_SM_PHASE3_SEND_SIG:
        cb->smState = BT_L2CAP_LINK_SM_PHASE3_IDLE;
        break;
    case BT_L2CAP_LINK_SM_PHASE3_IDLE:
        if (in != NULL) {
            switch(in->Code) {
            case SMP_Code_Encryption_Information:
                DEBUGP("SM: Encryption_Information recv\n");
                DUMP("key=", in->Encryption_Information.Long_Term_Key);
                break;
            case SMP_Code_Master_Identification:
                DEBUGP("SM: Master_Identification recv\n");
                DUMP("EDIV=", in->Master_Identification.EDIV);
                DUMP("Rand=", in->Master_Identification.Rand);
                break;
            case SMP_Code_Identity_Information:
                DEBUGP("SM: Identity_Information recv\n");
                DUMP("id=", in->Identity_Information.Identity_Resolving_Key);
                break;
            case SMP_Code_Identity_Address_Information:
                DEBUGP("SM: Identity_Address_Information recv type=%d\n", in->Identity_Address_Information.AddrType);
                DUMP("addr=", in->Identity_Address_Information.BD_ADDR);
                break;
            case SMP_Code_Signing_Information:
                DEBUGP("SM: Signing_Information recv\n");
                DUMP("key=", in->Signing_Information.Signature_Key);
                break;
            case SMP_Code_Pairing_Failed:
                DEBUGP("SM: Phase3 Faild!! Reason=0x%02x\n", in->Pairing_Failed.Reason);
                cb->smState = BT_L2CAP_LINK_SM_IDLE;
                break;
            default:
                DEBUGP("SM: Unsupport Code=0x%02x\n", in->Code);
                out->Code                  = SMP_Code_Pairing_Failed;
                out->Pairing_Failed.Reason = SMP_Reason_Command_Not_Supported;
                len = SMP_PDU_Format_HeadSize + sizeof(out->Pairing_Failed);
                break;
            }
        } else {
            if (!cb->encryption) {
                cb->smState = BT_L2CAP_LINK_SM_IDLE;
            }
        }
        break;

    default:
        DEBUGP("SM: bad state=%d\n",cb->smState);
        cb->smState = BT_L2CAP_LINK_SM_IDLE;
        break;
    }

    return(len);
}
