//
//  usb_host_bluetooth.c
//
//  USB Host Protocol Stack - USB Wireless controller class for bluetooth dongle
//  for Microchip MPLAB harmony
//  Copyright (C) 2015 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//              [Vol 4] Host Controller Interface [Transport Layer]
//              Part B: USB Transport Layer
//      Microchip
//          MPLAB Harmony Integrated Software Framework v1.01
//              Help document "help_harmony.pdf", Section "USB Host Library"
//
//  Device: PIC32MX
//      IDE     : MPLAB X IDE v2.26
//      Compiler: Xc32-gcc(v1.33)
//      Library : Microchip Harmony V1.01
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2015.01.29 v0.0  First cording
//  2015.02.03 v0.3  Commit
//  2015.03.03 v0.3b Concealment of bluetooth task to the user
//
#include "system_config.h"
#include "usb/usb_host.h"
#include "usb/src/usb_host_local.h"
#include "usb/usb_host_bluetooth.h"

#include "bt_hci.h"
#include "bt_process.h"                 // ADD 2015.03.03

//
//  Instance
//
typedef enum {
    IDLE = 0,
    POLL,
    COMMAND_START,
    COMMAND_POST,
    COMMAND_WAIT,
    COMMAND_END,
    EVENT_START,
    EVENT_POLL,
    EVENT_POLL_WAIT,
    EVENT_GET,
    EVENT_POST_WAIT,
    EVENT_END,
    ACL_OUT_START,
    ACL_OUT_POST,
    ACL_OUT_WAIT,
    ACL_OUT_END,
    ACL_IN_START,
    ACL_IN_POLL,
    ACL_IN_POLL_WAIT,
    ACL_IN_GET,
    ACL_IN_POST_WAIT,
    ACL_IN_END,
    LAST_STATE
} USB_HOST_BLUETOOTH_STATE;

typedef struct _USB_HOST_BLUETOOTH_INSTANCE_ {
    bool                        assigned;           // use or not
    uint8_t                     instanceNumber;
    uint8_t                     interfaceNumber;
    USB_HOST_DEVICE_ID          deviceId;           // hcid + device id
    USB_HOST_IRP                controlIRP;
    USB_HOST_IRP                interruptInIRP;
    USB_HOST_IRP                bulkInIRP;
    USB_HOST_IRP                bulkOutIRP;
    USB_HOST_PIPE_HANDLE        controlPipeHandle;
    USB_HOST_PIPE_HANDLE        interruptInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkOutPipeHandle;
    uint8_t                     interruptInEndpoint;
    uint8_t                     bulkInEndpoint;
    uint8_t                     bulkOutEndpoint;
    USB_SETUP_PACKET            setupPacket;
    USB_HOST_BLUETOOTH_EVENT_HANDLER  appEventCallBack;

    USB_HOST_BLUETOOTH_STATE    state;              //
    long                        cmdTryCount;        //
    long                        evtTryCount;        //
    long                        outTryCount;        //
    long                        inTryCount;         //
    size_t                      cmdBufLen;          //
    size_t                      evtBufLen;          //
    size_t                      outBufLen;          //
    size_t                      inBufLen;           //
    unsigned char               cmdBuf[64];         // USB endpoint buffer
    unsigned char               evtBuf[16];         //
    unsigned char               outBuf[64];         //
    unsigned char               inBuf[64];          //
} USB_HOST_BLUETOOTH_INSTANCE;

USB_HOST_BLUETOOTH_INSTANCE  gUSBHostBluetoothInstance[USB_HOST_BLUETOOTH_INSTANCES_NUMBER];


//
//  USB Driver(IRP) call back routines
//
void _USB_HOST_BLUETOOTH_ControlTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_BulkInTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_BulkOutTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_InterruptInTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

//
//  USB host class driver call back routines
//
USB_ERROR _USB_HOST_BLUETOOTH_DeviceDriverInitialize(USB_HOST_DEVICE_ID deviceId,
                                               uint8_t           *classInstance,
                                               USB_SPEED          speed)
{
    USB_HOST_QUERY              query;
    USB_ERROR                   status = USB_ERROR_NONE;
    USB_INTERFACE_DESCRIPTOR   *ifDesc;
    USB_ENDPOINT_DESCRIPTOR    *intr;
    USB_ENDPOINT_DESCRIPTOR    *bulkIn;
    USB_ENDPOINT_DESCRIPTOR    *bulkOut;
    USB_ENDPOINT_DESCRIPTOR     cntl;
    USB_HOST_BLUETOOTH_INSTANCE *p;
    uint8_t                     i;
    uint8_t                     interfaceId;

    interfaceId = 0;                    // Fixed
    _USB_HOST_SelectConfiguration(deviceId, 1);

    /* Find the free space for bluetooth instance */
    for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i ++) {
        p = &(gUSBHostBluetoothInstance[i]);
        if (p->assigned == false) {
            break;  /* Find free slot */
        }
    }

    /* validate the instance */
    if (i >= USB_HOST_BLUETOOTH_INSTANCES_NUMBER) {
        return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }

    *classInstance     = i;             /* count is the instance number return to host */
    p->assigned        = true;          /* Mark flag as use */
    p->deviceId        = deviceId;      /* store the Device address in instanceInfo */
    p->instanceNumber  = i;             /* save instance number */
    p->interfaceNumber = interfaceId;   /* bluetooth interface Number */

    /* Find interface */
    query.id                     = deviceId;
    query.class                  = USB_HOST_BLUETOOTH_CLASS;
    query.subClass               = USB_HOST_BLUETOOTH_SUBCLASS;
    query.protocol               = USB_HOST_BLUETOOTH_PROTOCOL;
    query.flags                  = USB_HOST_MATCH_CLASS_SUBCLASS_PROTOCOL;
    query.interfaceNumber        = interfaceId;
    query.alternateSettingNumber = 0x00;
    status = _USB_HOST_FindInterface(&query, &ifDesc);
    if (status != USB_ERROR_NONE) {
        return(USB_ERROR_HOST_INTERFACE_NOT_FOUND);
    }

    //
    //  Control pipe
    //
    /* Get control pipe information */
    cntl.transferType = USB_TRANSFER_TYPE_CONTROL;
    p->controlPipeHandle = _USB_HOST_PipeSetup(deviceId, &cntl, speed);
    if (p->controlPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
         return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }
    p->controlIRP.callback = _USB_HOST_BLUETOOTH_ControlTransferComplete;
    p->controlIRP.userData = (uintptr_t )p;

    //
    //  Interrupt pipe
    //
    query.endpointAddress   = 0 ; // we have to get temp
    query.endpointType      = USB_TRANSFER_TYPE_INTERRUPT;
    query.endpointDirection = USB_DATA_DIRECTION_DEVICE_TO_HOST;
    query.flags             = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &intr);
    if(status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->interruptInEndpoint   = intr->bEndpointAddress;
    p->interruptInPipeHandle = _USB_HOST_PipeSetup(deviceId, intr, speed);    /* Create Interrupt pipe */
    if(p->interruptInPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->interruptInIRP.callback = _USB_HOST_BLUETOOTH_InterruptInTransferComplete;
    p->interruptInIRP.size     = intr->wMaxPacketSize;
    p->interruptInIRP.userData = (uintptr_t)p;

    //
    //  Bulk in pipe
    //
    query.alternateSettingNumber = ifDesc->bAlternateSetting;
    query.endpointType           = USB_TRANSFER_TYPE_BULK;
    query.endpointDirection      = USB_DATA_DIRECTION_DEVICE_TO_HOST ;
    query.flags                  = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &bulkIn);
    if (status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->bulkInEndpoint   = bulkIn->bEndpointAddress;
    p->bulkInPipeHandle = _USB_HOST_PipeSetup(deviceId, bulkIn, speed);    /* Create pipe */
    if (p->bulkInPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->bulkInIRP.callback = _USB_HOST_BLUETOOTH_BulkInTransferComplete;
    p->bulkInIRP.size     = bulkIn->wMaxPacketSize ;
    p->bulkInIRP.userData = (uintptr_t)p;

    //
    //  Bulk out pipe
    //
    query.alternateSettingNumber = ifDesc->bAlternateSetting;
    query.endpointType           = USB_TRANSFER_TYPE_BULK;
    query.endpointDirection      = USB_DATA_DIRECTION_HOST_TO_DEVICE;
    query.flags                  = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &bulkOut);
    if (status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->bulkOutEndpoint   = bulkOut->bEndpointAddress;
    p->bulkOutPipeHandle = _USB_HOST_PipeSetup(deviceId, bulkOut, speed);    /* Create pipe */
    if (p->bulkOutPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->bulkOutIRP.callback = _USB_HOST_BLUETOOTH_BulkOutTransferComplete;
    p->bulkOutIRP.size     = bulkOut->wMaxPacketSize ;
    p->bulkOutIRP.userData = (uintptr_t )p;

    //
    //  Others
    //
    bt_setAliveHCI();		        // Bluetoot stack connection
    bt_init();                          // ADD 2015.03.03
    p->state = IDLE;
    p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ATTACH, NULL);
    return status;
}


void _USB_HOST_BLUETOOTH_DeInitialize(USB_HOST_DEVICE_ID id)
{
    USB_HOST_BLUETOOTH_INSTANCE  *p;
    uint8_t                 i;

    for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i++) {
        p = &(gUSBHostBluetoothInstance[i]);
        if (p->deviceId == id) {
            break;
        }
    }
    if (i >= USB_HOST_BLUETOOTH_INSTANCES_NUMBER) {
        return;
    }

    /* inform to the bluetooth device detach event to the application */
    p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_DETACH, NULL);
    p->state = IDLE ;
    bt_deinit();                        // ADD 2015.03.03
    bt_setDeadHCI();                    // Bluetoot stack connection

    _USB_HOST_PipeClose(p->bulkOutPipeHandle);
    _USB_HOST_PipeClose(p->bulkInPipeHandle);
    _USB_HOST_PipeClose(p->interruptInPipeHandle);

    /* close the instance */
    memset(p, 0, sizeof(USB_HOST_BLUETOOTH_INSTANCE));
    p->assigned = false;

    return;
}


USB_HOST_BLUETOOTH_HANDLE USB_HOST_BLUETOOTH_Open(uint8_t instance)
{
    USB_HOST_BLUETOOTH_INSTANCE  *p;
    USB_HOST_BLUETOOTH_HANDLE     h;

    /* Get instance driver information */
    p = &(gUSBHostBluetoothInstance[instance]);
    if(p->assigned == 0x01) {
        h = (USB_HOST_BLUETOOTH_HANDLE)p;
    } else {
        h = NULL;
    }
    return(h);
}


void USB_HOST_BLUETOOTH_Close(USB_HOST_BLUETOOTH_HANDLE h)
{
    //Invalidate the handle
    h = NULL;
    return;
}


USB_ERROR USB_HOST_BLUETOOTH_EventHandlerSet(USB_HOST_BLUETOOTH_EVENT_HANDLER handler)
{
    USB_HOST_BLUETOOTH_INDEX     i;
    USB_HOST_BLUETOOTH_INSTANCE *p;

    /* check if the given event handler is valid */
    if(handler) {
        for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i++) {
            p = &(gUSBHostBluetoothInstance[i]);
            /* update the event handler for this instance */
            p->appEventCallBack = handler ;
        }
        return(USB_ERROR_NONE);
    } else {
        return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }
}


void _USB_HOST_BLUETOOTH_Tasks(uint8_t instanceNumber)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    int                         rc;

    p = &(gUSBHostBluetoothInstance[instanceNumber]);
    if (p->assigned == false) {
        return;
    }

    // State machine
    switch(p->state){
    case IDLE:
        if (p->controlPipeHandle) {     // Pipe open?
            bt_setAliveHCI();           // Bluetoot stack connection
            p->interruptInIRP.flags = 0;
            p->interruptInIRP.data  = p->evtBuf;
            p->interruptInIRP.size  = sizeof(p->evtBuf);
            _USB_HOST_IRPSubmit(p->interruptInPipeHandle, &(p->interruptInIRP));
            p->bulkInIRP.flags = 0;
            p->bulkInIRP.data  = p->inBuf;
            p->bulkInIRP.size  = sizeof(p->inBuf);
            _USB_HOST_IRPSubmit(p->bulkInPipeHandle, &(p->bulkInIRP));
            p->state = POLL;
        } else {
            bt_setDeadHCI();            // Bluetoot stack connection
        }
        break;

    case POLL:
        if ((rc = bt_getHCIcommand(p->cmdBuf, sizeof(p->cmdBuf))) > 0) { // COMMAND
            p->cmdTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->cmdBufLen   = rc;
            p->state       = COMMAND_POST;
        } else if (p->interruptInIRP.status == USB_HOST_IRP_STATUS_COMPLETED) { // EVENT
            p->evtBufLen   = p->interruptInIRP.size;
            p->state       = EVENT_POST_WAIT;
        } else if (p->interruptInIRP.status < 0) {
            p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
            p->evtTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state       = IDLE;
        } else if ((rc = bt_getACLout(p->outBuf, sizeof(p->outBuf))) > 0) { // ACL OUT
            p->outTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->outBufLen   = rc;
            p->state       = ACL_OUT_POST;
        } else if (p->bulkInIRP.status == USB_HOST_IRP_STATUS_COMPLETED) { // ACL IN
            p->inBufLen    = p->bulkInIRP.size;
            p->state       = ACL_IN_POST_WAIT;
        } else if (p->bulkInIRP.status < 0) {
            p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
            p->inTryCount  = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state       = IDLE;
        }
        break;

    //---------------------------------------------------------------------
    case COMMAND_POST:
        p->setupPacket.bmRequestType = USB_SETUP_DIRN_HOST_TO_DEVICE |
                                       USB_SETUP_TYPE_CLASS          |
                                       USB_SETUP_RECIPIENT_DEVICE;
        p->setupPacket.bRequest      = 0;
        p->setupPacket.wValue        = 0;
        p->setupPacket.wIndex        = 0;
        p->setupPacket.wLength       = p->cmdBufLen;
        p->controlIRP.setup = &(p->setupPacket);
        p->controlIRP.flags = 0;
        p->controlIRP.data  = p->cmdBuf;
        p->controlIRP.size  = p->cmdBufLen;
        _USB_HOST_IRPSubmit(p->controlPipeHandle, &(p->controlIRP));
        p->state = COMMAND_WAIT;
        break;

    case COMMAND_WAIT:
        if (p->controlIRP.status == USB_HOST_IRP_STATUS_COMPLETED) {
            p->state = POLL;
        } else if (p->controlIRP.status < 0) {
            if (--p->cmdTryCount <= 0) {
//              status   = USBH_UNRECOVERED_ERROR;
                p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case EVENT_POST_WAIT:
        if (bt_putHCIevent(p->evtBuf, p->evtBufLen) == p->evtBufLen) {
            p->interruptInIRP.flags = 0;    // Next message
            p->interruptInIRP.data  = p->evtBuf;
            p->interruptInIRP.size  = sizeof(p->evtBuf);
            _USB_HOST_IRPSubmit(p->interruptInPipeHandle, &(p->interruptInIRP));
            p->state = POLL;
        }
        break;

    //---------------------------------------------------------------------
    case ACL_OUT_POST:
        p->bulkOutIRP.flags = 0;
        p->bulkOutIRP.data  = p->outBuf;
        p->bulkOutIRP.size  = p->outBufLen;
        _USB_HOST_IRPSubmit(p->bulkOutPipeHandle, &(p->bulkOutIRP));
        p->state = ACL_OUT_WAIT;
        break;

    case ACL_OUT_WAIT:
        if (p->bulkOutIRP.status == USB_HOST_IRP_STATUS_COMPLETED) {
            p->state = POLL;
        } else if (p->bulkOutIRP.status < 0) {  // ERROR
            if (--p->outTryCount <= 0) {
//              status   = USBH_UNRECOVERED_ERROR;
                p->appEventCallBack(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case ACL_IN_POST_WAIT:
        if (bt_putHCIin(p->inBuf, p->inBufLen) == p->inBufLen) {
            p->bulkInIRP.flags = 0;
            p->bulkInIRP.data  = p->inBuf;
            p->bulkInIRP.size  = sizeof(p->inBuf);
            _USB_HOST_IRPSubmit(p->bulkInPipeHandle, &(p->bulkInIRP));
            p->inTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state      = POLL;
        }
        break;

    default:
        p->state = IDLE;
        break;
    }

    bt_process();                       // ADD 2015.03.03

    return;
}

//
//  USB host class driver jump table
//
USB_HOST_CLASS_DRIVER usbHostClassDriverBluetooth = {
    .initializeDeviceDriver    = _USB_HOST_BLUETOOTH_DeviceDriverInitialize,    /* bluetooth device level init function */
    .initializeInterfaceDriver = NULL,                                          /* bluetooth interface level init function */
    .deInitializeDriver        = _USB_HOST_BLUETOOTH_DeInitialize,              /* bluetooth de-init function */
    .task                      = _USB_HOST_BLUETOOTH_Tasks                      /* bluetooth tasks function */
};
